/*
** Copyright (C) 2003-2006 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

#include "silk.h"

RCSIDENT("$SiLK: flowcap.c 5934 2006-12-12 19:34:46Z mthomas $");

#include <glob.h>
#include <sys/un.h>

#include "flowcap.h"
#include "utils.h"
#include "fcinternals.h"
#include "sksite.h"
#include "skdeque.h"
#include "sklog.h"
#include "skdaemon.h"
#include "sktransfer.h"


/* TYPEDEFS AND DEFINES */

/* Where to print --help output */
#define USAGE_FH stdout

/*
 *  For debugging.  Argument should be surrounded by double-parens:
 *      LOGDEBUG(("My value is %d", value));
 */
#ifdef DEBUG
#  define LOGDEBUG(args) DEBUGMSG args
#endif
#ifndef LOGDEBUG
#  define LOGDEBUG(args)
#endif

/* It seems that solaris does not define SUN_LEN (yet) */
#ifndef SUN_LEN
#define SUN_LEN(su) \
        (sizeof(*(su)) - sizeof((su)->sun_path) + strlen((su)->sun_path))
#endif


/* Max timestamp length (YYYYMMDDhhmmss) */
#define FC_TIMESTAMP_MAX 15
/* Maximum sensor size (including either trailing zero or preceeding hyphen) */
#define FC_SENSOR_MAX (SK_MAX_STRLEN_SENSOR + 1)
/* Maximum probe size (including either trailing zero or preceeding hyphen) */
#define FC_PROBE_MAX (SK_MAX_STRLEN_SENSOR + 1)
/* Size of uniquness extension */
#define FC_UNIQUE_MAX 7
/* Previous two, plus hyphen */
#define FC_NAME_MAX                                     \
    (FC_TIMESTAMP_MAX + FC_SENSOR_MAX +                 \
     FC_PROBE_MAX + FC_UNIQUE_MAX)


/* Minimum flowcap version */
/* We no longer support flowcap version 1 */
#define FC_VERSION_MIN 2

/* Default version of flowcap to produce */
#define FC_VERSION_DEFAULT 5

/* Indexes for DISK and RAM related pairs. */
#define FC_DISK 0
#define FC_RAM 1


/* LOCAL FUNCTIONS */

static void appUsageLong(void);/* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);/* never returns */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static void parseSensorList(const char *sensor_list);
static int optionsDirCheck(const char *opt_name, const char *opt_arg);
static void loadFiles(void);
static void createWriteFiles(void);
static skTimerRepeat_t timerHandler(void *t);
static void addToLowPriorityQueue(fileStructPtr fptr);
static void addToHighPriorityQueue(fileStructPtr fptr);
static fileStructPtr getFileStruct(void);
static void addToFreeQueue(fileStructPtr fPtr);
static void unloadFiles(void);
static void openWFile(writeFileStructPtr wfPtr);
static void openWFileBase(writeFileStructPtr wfPtr);
static int  startReader(void);
static int  startSender(void);
static void stopReader(void);
static void stopSender(void);
static void shutdownInterface(void *);
static void restartInterface(void *);
static void dumpStats(fileStructPtr fPtr);
static void initQueues(void);
static void destroyQueues(void);
static void closeWFileBase(writeFileStructPtr wfPtr, uint8_t open);


/* EXPORTED VARIABLES */

/* Flag used to stop readers.  When readers notice that this has gone
 * from 1 to 0, they know to shut down.  */
volatile uint8_t reading = 0;


/* LOCAL VARIABLES */

/* Send queue */
static skDeque_t sendQ;

/* Send queues split into priorites*/
static skDeque_t lowQ, highQ;

/* Queue in which no longer needed fileStructs can be recycled. */
static skDeque_t freeQ;

/* File which is currently being sent */
static fileStructPtr curSendFPtr;

/* Amount of disk space allocated for flowcap files {DISK, RAM} */
static uint64_t maxSpace[2];

/* Amount of disk space which has not already been allocated to files
   {DISK, RAM} */
static uint64_t unallocated_space[2];

/* Mutex and condition variable to protect the unallocated_space
   variable.  This allows threads to wait (on spaceCond) until there
   is enough unallocated space.  */
static pthread_mutex_t spaceMutex;
static pthread_cond_t spaceCond;

/* Indicator of whether flowcap is in the process of shutting down */
static volatile int shuttingDown = 0;

/* High (RAM) and low (DISK) speed disk locations {DISK, RAM} */
static char *destDirs[2] = {NULL, NULL};

/* File write timeout in seconds.  A file will be closed out if this
   amount of time passes without getting too large. */
static uint32_t writeTimeout = 60;

/* ACK timout in seconds. */
static uint32_t ackTimeout = 10;

/* Address of the flowcap client (consumer). */
static in_addr_t fc_client_addr;

/* Port upon which flowcap listens for connecting clients (consumers). */
static uint16_t fc_listen_port;

/* Files will be split into multiple files if they reach this size. */
static uint32_t maxFileSize;

/* Ammount of disk space to allocate to a file before we know its
   actual size. */
static uint64_t allocFileSize;

/* Main thread id. */
static pthread_t main_thread;

/* Whether the interface is currently up.  (Boolean) */
static uint8_t interfaceup = 1;

/* Whether the probe configuration file has been parsed (Boolean) */
static uint8_t probes_loaded;

/* The version of flowcap to produce */
static uint8_t flowcap_version = FC_VERSION_DEFAULT;

/* The file sender. */
static skTransfer_t sender;

/* The array of write file structures, and the array's size. */
static writeFileStructPtr write_files;
static size_t num_write_files;

/* The list of probes we care about */
static sk_vector_t *probe_vec = NULL;

/* Stashed sensor list */
static char *sensor_list;


/* OPTIONS SETUP */

typedef enum {
    OPT_SENSOR_CONFIG, OPT_PORT, OPT_CLIENT_ADDRESS,
    OPT_DISK_DIRECTORY, OPT_DISK_SPACE, OPT_MAX_FILE_SIZE,
    OPT_RAM_DIRECTORY, OPT_RAM_SPACE,
    OPT_TIMEOUT, OPT_ACKTIMEOUT, OPT_SENSORS,
    OPT_VERSION
} appOptionsEnum;

static struct option appOptions[] = {
    {"sensor-configuration", REQUIRED_ARG, 0, OPT_SENSOR_CONFIG},
    {"fc-port",              REQUIRED_ARG, 0, OPT_PORT},
    {"client-address",       REQUIRED_ARG, 0, OPT_CLIENT_ADDRESS},
    {"disk-directory",       REQUIRED_ARG, 0, OPT_DISK_DIRECTORY},
    {"disk-space",           REQUIRED_ARG, 0, OPT_DISK_SPACE},
    {"max-file-size",        REQUIRED_ARG, 0, OPT_MAX_FILE_SIZE},
    {"ram-directory",        REQUIRED_ARG, 0, OPT_RAM_DIRECTORY},
    {"ram-space",            REQUIRED_ARG, 0, OPT_RAM_SPACE},
    {"timeout",              REQUIRED_ARG, 0, OPT_TIMEOUT},
    {"ack-timeout",          REQUIRED_ARG, 0, OPT_ACKTIMEOUT},
    {"sensors",              REQUIRED_ARG, 0, OPT_SENSORS},
    {"fc-version",           REQUIRED_ARG, 0, OPT_VERSION},
    {0,0,0,0}                /* sentinel entry */
};

static char *appHelp[] = {
    "Sensor configuration file.",
    ("Port upon which flowcap listens for connections from\n"
     "\ta flowcap data-reading client."),
    ("Internet addresses which are permitted to connect\n"
     "\tto the flowcap port. (default: any)"),
    ("Directory in which to store aggregated packed flows\n"
     "\tbefore they are sent to the data-reading client."),
    ("Amount of space allotted for disk-based files, in bytes.\n"
     "\tAppend k, m, g, t for kilo-, mega-, giga-, tera-bytes, respectively"),
    ("Max file size, in bytes.  Accepts k,m,g,t suffix."),
    "Ram disk mount point. (default: none)",
    ("Amount of space allotted for ramdisk-based files.\n"
     "\tAccepts k,m,g,t suffix.  (default: 0)"),
    "Timeout for file writing, in seconds. (default: 60)",
    ("How long (in seconds) to wait for an ack from a\n"
     "\treceiver. (default: 10)"),
    "Sensor names from the configuration to use. (default: all)",
    NULL, /* generate dynamically */
    (char *)NULL
};


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH and exit the
 *    application normally.  The function is passed to optionsSetup()
 *    to print the usage when the --help option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                           \
    ("<SWITCHES>\n"                                                         \
     "\tflowcap is a daemon which listens to devices which produce flow\n"  \
     "\tdata (flow sources), homogenizes the data, stores it, and\n"        \
     "\tforwards as a compressed stream to a flowcap client program.\n")

    FILE *fh = USAGE_FH;
    int i;

    fprintf(fh, "%s %s", skAppName(), USAGE_MSG);
    fprintf(fh, "\nSWITCHES:\n");
    skOptionsDefaultUsage(fh);

    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        switch ((appOptionsEnum)i) {
          case OPT_VERSION:
            fprintf(fh,
                    "Version of flowcap protocol to use: %d-%d. (default: %d)",
                    FC_VERSION_MIN, FC_VERSIONS, flowcap_version);
            break;
          default:
            fprintf(fh, "%s", appHelp[i]);
            break;
        }
        fprintf(fh, "\n");
    }

    skdaemonOptionsUsage(fh);
    fprintf(fh, "\nSwitches with no defaults are required.\n");
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    Called by optionsParse(), this handles a user-specified switch
 *    that the application has registered, typically by setting global
 *    variables.  Returns 1 if the switch processing failed or 0 if it
 *    succeeded.  Returning a non-zero from from the handler causes
 *    optionsParse() to return a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    struct hostent *he;
    uint32_t tmp_number;
    uint64_t tmp_64;

    switch ((appOptionsEnum)opt_index) {
      case OPT_SENSOR_CONFIG:        /* probefile */
        if (probeConfParse(opt_arg) != 0) {
            skAppPrintErr("Error in probe configuration file.");
            return 1;
        }
        probes_loaded = 1;
        break;

      case OPT_CLIENT_ADDRESS:      /* client ip */
        if (((he = gethostbyname(opt_arg)) == NULL) ||
            he->h_length != sizeof (in_addr_t)) {
            skAppPrintErr("Illegal %s address '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        memcpy(&fc_client_addr, he->h_addr, sizeof (in_addr_t));
        break;

      case OPT_PORT:         /* fc port */
        if (skStringParseUint32(&tmp_number, opt_arg, 0, 0xFFFE)) {
            skAppPrintErr("Invalid argument to --%s switch: %s",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        fc_listen_port = (uint16_t)tmp_number;
        break;

      case OPT_RAM_DIRECTORY:          /* ram-directory */
        if (optionsDirCheck(appOptions[opt_index].name, opt_arg)) {
            return 1;
        }
        if (strlen(opt_arg) > PATH_MAX - FC_NAME_MAX) {
            skAppPrintErr("The %s name is too long '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        destDirs[FC_RAM] = strdup(opt_arg);
        break;

      case OPT_DISK_DIRECTORY:         /* disk-directory */
        if (optionsDirCheck(appOptions[opt_index].name, opt_arg)) {
            return 1;
        }
        if (strlen(opt_arg) > PATH_MAX - FC_NAME_MAX) {
            skAppPrintErr("The %s name is too long '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        destDirs[FC_DISK] = strdup(opt_arg);
        break;

      case OPT_MAX_FILE_SIZE:     /* max-file-size */
        if (skStringParseHumanUint64(&tmp_64, opt_arg, SK_HUMAN_NORMAL)) {
            skAppPrintErr("Invalid argument to --%s switch: %s",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        if (tmp_64 > 0xFFFFFFFF) {
            skAppPrintErr("Argument to --%s switch, %s, is too large",
                          appOptions[opt_index].name, opt_arg);
        }
        maxFileSize = tmp_64;
        break;

      case OPT_DISK_SPACE:       /* disk-space */
        if (skStringParseHumanUint64(&tmp_64, opt_arg, SK_HUMAN_NORMAL)) {
            skAppPrintErr("Invalid argument to --%s switch: %s",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        maxSpace[FC_DISK] = tmp_64;
        break;

      case OPT_RAM_SPACE:        /* ram-space */
        if (skStringParseHumanUint64(&tmp_64, opt_arg, SK_HUMAN_NORMAL)) {
            skAppPrintErr("Invalid argument to --%s switch: %s",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        maxSpace[FC_RAM] = tmp_64;
        break;

      case OPT_TIMEOUT:          /* timeout */
        if (skStringParseUint32(&tmp_number, opt_arg, 0, 0xFFFFFFFE)) {
            skAppPrintErr("Invalid argument to --%s switch: %s",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        writeTimeout = tmp_number;
        break;

      case OPT_ACKTIMEOUT:       /* ack-timeout */
        if (skStringParseUint32(&tmp_number, opt_arg, 0, 0xFFFFFFFE)) {
            skAppPrintErr("Invalid argument to --%s switch: %s",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        ackTimeout = tmp_number;
        break;

      case OPT_SENSORS:          /* sensors */
        sensor_list = opt_arg;
        break;

      case OPT_VERSION:          /* fc-version */
        if (skStringParseUint32(&tmp_number, opt_arg,
                                FC_VERSION_MIN, FC_VERSIONS))
        {
            skAppPrintErr(("Invalid argument to --%s switch: %s\n"
                           "\tValid arguments are 1-%d"),
                          appOptions[opt_index].name, opt_arg, FC_VERSIONS);
            return 1;
        }
        flowcap_version = tmp_number;
        break;

    } /* switch */

    return 0;
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
    static uint8_t teardownFlag = 0;

    pthread_mutex_lock(&mutex);

    if (teardownFlag) {
        pthread_mutex_unlock(&mutex);
        return;
    }
    teardownFlag = 1;
    pthread_mutex_unlock(&mutex);

    logMsg("Shutting down");
    shuttingDown = 1;
    optionsTeardown();
    stopReader();
    stopSender();
    logMsg("Unloading files");
    unloadFiles();

    destroyQueues();

    probeConfTeardown();
    skdaemonTeardown();

    return;
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int32_t argc, char **argv)
{
    int32_t nextOptIndex;

    skAppRegister(argv[0]);

    if (argc == 1) {
        skAppUsage();
    }

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* initialize the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);/* never returns */
    }

    /* register the options and the handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);
    }

    /* flowcap runs as a daemon */
    if (skdaemonSetup((SKLOG_FEATURE_LEGACY | SKLOG_FEATURE_SYSLOG),
                      argc, argv))
    {
        exit(EXIT_FAILURE);
    }

    if (probeConfSetup() != 0) {
        skAppPrintErr("unable to setup probe configuration handler");
        exit(EXIT_FAILURE);
    }

    /* zero out variables */
    fc_client_addr = INADDR_ANY;
    fc_listen_port = 0xFFFF;
    maxFileSize = 0;
    allocFileSize = 0;
    probes_loaded = 0;
    memset(maxSpace, 0, sizeof(maxSpace));
    memset(unallocated_space, 0, sizeof(unallocated_space));
    pthread_mutex_init(&spaceMutex, NULL);
    pthread_cond_init(&spaceCond, NULL);
    sensor_list = NULL;

    /* parse options. If OK, nextOptIndex is the first arg not used up */
    nextOptIndex = optionsParse(argc, argv);
    if (nextOptIndex < 0) {
        skAppUsage();/* never returns */
    }

    if (probes_loaded == 0) {
        skAppPrintErr("No probe definition file was specified");
        skAppUsage();
    }

    if (probeConfProbeCount() == 0) {
        skAppPrintErr("The probe definition file has no probes within it");
        exit(EXIT_FAILURE);
    }

    if (fc_listen_port == 0xFFFF) {
        skAppPrintErr("No flowcap port");
        skAppUsage();
    }

    if (!maxFileSize) {
        skAppPrintErr("No max-file-size");
        skAppUsage();
    }
    allocFileSize = maxFileSize + (LZO_MAX_COMP_SIZE - LZO_MAX_BUF_SIZE);

    if (maxSpace[FC_DISK] == 0 && maxSpace[FC_RAM] == 0) {
        skAppPrintErr("Need to include at least one of "
                      "disk-space or ram-space.");
        skAppUsage();
    }
    unallocated_space[FC_DISK] = maxSpace[FC_DISK];
    unallocated_space[FC_RAM] = maxSpace[FC_RAM];

    if (!destDirs[FC_DISK]) {
        skAppPrintErr("No disk-directory");
        skAppUsage();
    }

    if (maxSpace[FC_RAM] && !destDirs[FC_RAM]) {
        skAppPrintErr("No ram-directory");
        skAppUsage();
    }

    if (maxSpace[FC_RAM] > maxSpace[FC_DISK]) {
        skAppPrintErr("Ram files won't fit on disk");
        exit(EXIT_FAILURE);
    }
    maxSpace[FC_DISK] -= maxSpace[FC_RAM]; /* Adjust */

    /* verify the required options for logging */
    if (skdaemonOptionsVerify()) {
        exit(EXIT_FAILURE);
    }

    /* Check that there aren't any extraneous arguments */
    if (nextOptIndex != argc) {
        skAppPrintErr("too many or unrecognized argument specified '%s'",
                      argv[nextOptIndex]);
        skAppUsage();             /* never returns */
    }

    probe_vec = skVectorNew(sizeof(probe_def_t *));
    if (probe_vec == NULL) {
        skAppPrintErr("vector create failed");
        exit(EXIT_FAILURE);
    }

    if (sensor_list) {
        parseSensorList(sensor_list);
    } else {
        size_t i;

        for (i = 0; i < probeConfProbeCount(); i++)
        {
            const probe_def_t *probe = probeConfProbeGetByPosition(i);
            int rv = skVectorAppendValue(probe_vec, &probe);
            if (rv == -1) {
                skAppPrintErr("vector append failed");
                exit(EXIT_FAILURE);
            }
        }
    }

    if (skVectorGetCount(probe_vec) < 1) {
        skAppPrintErr("No sensors with probes were specified");
        exit(EXIT_FAILURE);
    }

    return;
} /* appSetup */



/* Parse a sensor list and add the probes associated with it into an
   already allocated probe_vec. */
static void parseSensorList(const char *sensor_list)
{
    sk_vector_t *sensor_vec;
    int count;
    int i;

    sensor_vec = skVectorNew(sizeof(sensorID_t));
    if (sensor_vec == NULL) {
        skAppPrintErr("vector create failed");
        exit(EXIT_FAILURE);
    }

    count = sksiteParseSensorList(sensor_vec, sensor_list);

    if (count < 1) {
        skAppPrintErr("Invalid argument to --%s switch: %s",
                      appOptions[OPT_SENSORS].name, sensor_list);
        exit(EXIT_FAILURE);
    }

    for (i = 0; i < count; i++) {
        sensorID_t sensor_id;
        int num_probes;

        skVectorGetValue(&sensor_id, sensor_vec, i);
        num_probes = probeConfGetProbesForSensor(probe_vec, sensor_id);
        if (num_probes == -1) {
            char s_name[SK_MAX_STRLEN_SENSOR+1];
            skAppPrintErr("Error adding probes for %s.",
                          sksiteSensorGetNameSafe(s_name, sensor_id,
                                                  sizeof(s_name)));
            exit(EXIT_FAILURE);
        }
    }

    skVectorDestroy(sensor_vec);
}


/*
 *  dir_exists = optionsDirCheck(opt_name, opt_arg);
 *
 *    Verify that the directory in 'opt_arg' exists and that we have a
 *    full path to the directory.  If so, return 0; otherwise, print
 *    an error that the option named by 'opt_name' was bad and return
 *    -1;
 */
static int optionsDirCheck(const char *opt_name, const char *opt_arg)
{
    if (!opt_arg || !opt_arg[0]) {
        skAppPrintErr("The directory argument to %s is empty",
                      opt_name);
        return -1;
    }

    if (!dirExists(opt_arg)) {
        skAppPrintErr("The %s '%s' does not exist",
                      opt_name, opt_arg);
        return -1;
    }

    if (opt_arg[0] != '/') {
        skAppPrintErr(("Must use complete path to %s.\n"
                       "\t('%s' does not begin with slash)"),
                      opt_name, opt_arg);
        return -1;
    }

    return 0;
}


/*
 * loadFiles:
 *      Load files from real disk into their data structures renaming
 *      as we go along.  Then enqueue them for sending.
 *      Finally, grab files from the free queue to open writers.
 */
static void loadFiles(void)
{
    fileStructPtr fPtr;
    uint32_t i;
    char newName[PATH_MAX];
    char *fileName;
    glob_t files;
    int rv;

    /* Empty the ramdir */
    if (destDirs[FC_RAM]) {

        sprintf(newName, "%s/*", destDirs[FC_RAM]);
        if ((rv = glob(newName, GLOB_ERR, NULL, &files)) != 0
#if HAVE_DECL_GLOB_NOMATCH
            && rv != GLOB_NOMATCH
#endif
            )
        {
            skAppPrintErr("Error doing globbing on %s", newName);
            perror(NULL);
            exit(EXIT_FAILURE);
        }
#if HAVE_DECL_GLOB_NOMATCH
        if (rv == GLOB_NOMATCH) {
            files.gl_pathc = 0;
        }
#endif
        for (i = 0; i < (uint32_t)files.gl_pathc; i++) {
            fileName = files.gl_pathv[i];
            if (fileName[0] == '.') {
                continue;
            }
            DEBUGMSG("%s:%d Unlinking %s", __FILE__, __LINE__, fileName);
            unlink(fileName);
        }
        globfree(&files);
    }

    /* Get a list of current files from the disk dir*/
    sprintf(newName, "%s/*", destDirs[FC_DISK]);
    if ((rv = glob(newName, GLOB_ERR, NULL, &files)) != 0
#if HAVE_DECL_GLOB_NOMATCH
        && rv != GLOB_NOMATCH
#endif
        )
    {
        skAppPrintErr("Error doing globbing on %s", newName);
        perror(NULL);
        exit(EXIT_FAILURE);
    }
#if HAVE_DECL_GLOB_NOMATCH
    if (rv == GLOB_NOMATCH) {
        files.gl_pathc = 0;
    }
#endif
    /* Add these files to the send queues */
    for (i = 0; i < (uint32_t)files.gl_pathc; i++) {
        const probe_def_t *found = NULL;
        const char *sensor_start, *sensor_end;
        const char *probe_start, *probe_end;
        const char *name_start;

        fileName = files.gl_pathv[i];

        /* Skip directory special files */
        if (fileName[0] == '.') {
            continue;
        }

        /* If the file is empty (header only or smaller), delete.  The
           extra four bytes are for the uint32_t zero which is added
           for an empty compressed block. */
        if (fileSize(fileName) <= (off_t)(sizeof(fcHeader_V2_t) + 4)) {
            DEBUGMSG("%s:%d Unlinking %s", __FILE__, __LINE__, fileName);
            unlink(fileName);
            LOGDEBUG(("Deleting %s", fileName));
            continue;
        }

        name_start = strrchr(fileName, '/');
        if (name_start == NULL) {
            name_start = fileName;
        } else {
            name_start++;
        }

        /* Allocate the fPtr structure */
        fPtr = (fileStructPtr) calloc(sizeof(fileStruct), 1);
        fPtr->startTime = fPtr->endTime = (uint32_t)time(NULL);
        fPtr->size = fileSize(fileName);
        snprintf(fPtr->fName, sizeof(fPtr->fName), "%s", name_start);

        /* This isn't protected because it should run before the other
           main flowcap threads are created. */
        unallocated_space[FC_DISK] -= fPtr->size;

        /* Here we parse the sensor name and probe name from the
           filename */
        if ((sensor_start = strchr(name_start, '_'))) {
            sensor_start++;
            if (*sensor_start != '\0' &&
                (sensor_end = strchr(sensor_start, '_')))
            {
                probe_start = sensor_end + 1;
                if (*probe_start != '\0' &&
                    (probe_end = strchr(probe_start, '.')))
                {
                    /* Here, the sensor name and probe name have been
                       successfully parsed, so we now look for a probe
                       match. */
                    char sensor_nm[FC_SENSOR_MAX];
                    char probe_nm[FC_PROBE_MAX];

                    /* Copy the names */
                    strncpy(sensor_nm, sensor_start,
                            sensor_end - sensor_start);
                    sensor_nm[sensor_end - sensor_start] = 0;

                    strncpy(probe_nm, probe_start,
                            probe_end - probe_start);
                    probe_nm[probe_end - probe_start] = 0;

                    /* Find the probe */
                    found = probeConfGetProbeByName(sensor_nm, probe_nm);
                }
            }
            /* Add to the appropriate queue.  Unknown is assumed to be
               high priority. */
            if (!found || skProbeGetPriority(found) > 50) {
                addToHighPriorityQueue(fPtr);
            } else {
                addToLowPriorityQueue(fPtr);
            }
        }

    }
    globfree(&files);
    return;
}


/*
 * createWriteFiles:
 *     Creates the write file structures
 */
static void createWriteFiles(void)
{
    size_t i;

    num_write_files = skVectorGetCount(probe_vec);

    write_files = (writeFileStructPtr)calloc(sizeof(writeFileStruct),
                                             num_write_files);

    for (i = 0; i < num_write_files; i++) {
        probe_def_t *probe;

        skVectorGetValue(&probe, probe_vec, i);

        /* Fill in the probe */
        write_files[i].probe = probe;
        write_files[i].type = FT_FLOWCAP;

        /* The rest of the initialization */
        write_files[i].version = flowcap_version;
        write_files[i].HWM = maxFileSize;
        write_files[i].lzofile = lzo_create_compr_buffer();
        pthread_mutex_init(&write_files[i].mutex, NULL);

        /* Create the file */
        openWFile(&write_files[i]);
    }
}


/*
 * timerHandler:
 *      Timer fired for a file. Ship it.
 * IMPORTANT:  We must set the struct's timerHandle to NULL since,
 *      otherwise, closeWFile() will free it and then the selector
 *      lib will free it again.
 */
static skTimerRepeat_t timerHandler(void *t)
{
    register writeFileStructPtr wfPtr = (writeFileStructPtr)t;

    /* Set the close flag first. */
    wfPtr->close = 1;

    /* Timer handler stuff */
    logMsg("Timer fired for %s", wfPtr->fPtr->fName);
    wfPtr->timerHandle = NULL;

    /* Close the file, and open a new one. */
    closeWFile(wfPtr, 1);

    return SK_TIMER_END;
}



/*
 * addToLowPriorityQueue:
 *      Add a fileStructPtr to the acl send queue.
 *      If possible, initiate another file send.
 */
static void addToLowPriorityQueue(fileStructPtr fPtr)
{
    LOGDEBUG(("Adding %s to acl send queue", fPtr->fName));
    skDequePushFront(lowQ, (void *)fPtr);

    return;
}


/* called by sender when sending failed */
static void addToSendQueue(fileStructPtr fPtr)
{
    skDequePushFront(sendQ, (void *)fPtr);
}


/*
 * addToHighPriorityQueue:
 *      Add a fileStructPtr to the routed send queue.
 *      If possible, initiate another file send.
 */
static void addToHighPriorityQueue(fileStructPtr fPtr)
{
    LOGDEBUG(("Adding %s to routed send queue", fPtr->fName));
    skDequePushFront(highQ, (void *)fPtr);

    return;
}


static void openWFile(writeFileStructPtr wfPtr)
{
    LOGDEBUG(("Locking in openWfile."));
    pthread_cleanup_push((cleanupHandler)pthread_mutex_unlock,
                         (void *)&wfPtr->mutex);
    pthread_mutex_lock(&wfPtr->mutex);
    LOGDEBUG(("Locked in openWfile."));

    openWFileBase(wfPtr);

    LOGDEBUG(("Unlocking in openWfile."));
    pthread_cleanup_pop(1);
    LOGDEBUG(("Unlocked in openWfile."));
}

/*
 * openWFileBase
 *      Open the next available file as a write file.
 *      Take care of file naming, cycling the global file seqnum
 *      Open file with gzopen.
 *      Stop the reader if no files are available.
 *      If we are actively reading, install a timeout.
 *      Write the appropriate header depending on wfPtr->type.
 */
static void openWFileBase(writeFileStructPtr wfPtr)
{
    char buffer[PATH_MAX];
    char ts[FC_TIMESTAMP_MAX + 1]; /* timestamp */
    time_t ct;
    struct tm ut;
    int fd;

    logMsg("Opening new file.");

    /* Get an empty fptr */
    wfPtr->fPtr = getFileStruct();

    assert(flowcap_version >= FC_VERSION_MIN);
    assert(flowcap_version <= FC_VERSIONS);
    assert(FC_REC_SIZE(flowcap_version) != -1);

    /* Create a timestamp */
    ct = time(NULL);
    gmtime_r(&ct, &ut);
    strftime(ts, sizeof(ts), "%Y%m%d%H%M%S", &ut);

    /* Create a filename from the timestamp and probe */
    if (snprintf(wfPtr->fPtr->fName, sizeof(wfPtr->fPtr->fName),
                 "%s_%s_%s.XXXXXX", ts,
                 skProbeGetSensorName(wfPtr->probe),
                 skProbeGetName(wfPtr->probe)) >=
        (int)sizeof(wfPtr->fPtr->fName))
    {
        logMsg("Filename exceeded transfer filename size.");
        exit(EXIT_FAILURE);
    }


    /* Add in the directory */
    if (snprintf(buffer, sizeof(buffer), "%s/%s",
                 destDirs[wfPtr->fPtr->isRamDisk],
                 wfPtr->fPtr->fName) >= (int)sizeof(buffer)) {
        logMsg("Filename exceeded buffer size.");
        exit(EXIT_FAILURE);
    }

    /* Open the file; making sure it is unique */
    fd = mkstemp(buffer);
    if (fd == -1) {
        logMsg("Write file creation (%s) error: %s", buffer, strerror(errno));
        exit(EXIT_FAILURE);
    }

    /* Put the uniquified name back in the fptr */
    baseName_r(wfPtr->fPtr->fName, buffer, sizeof(wfPtr->fPtr->fName));

    DEBUGMSG("File path: %s", buffer);
    DEBUGMSG("File name: %s", wfPtr->fPtr->fName);

    /* Set the permissions */
    fchmod(fd, 0644);

    /* Get a FILE * handle to the file */
    if ((wfPtr->gzFILE = fdopen(fd, "w")) == NULL) {
        logMsg("Write file open (%s) error: %s", buffer, strerror(errno));
        exit(EXIT_FAILURE);
    }

    LOGDEBUG(("%s opened for write", buffer));

    /* write the header, and retrieve the size of the header */
    wfPtr->fPtr->size =
        wfPtr->fPtr->uncompressSize =
        wfPtr->headerSize = dumpHeader(wfPtr);

    /* Bind the compression buffer to the file */
    lzo_bind_compr_buffer(wfPtr->lzofile, wfPtr->gzFILE);

    /* Set up default values */
    wfPtr->fPtr->startTime = time((time_t *)NULL);
    wfPtr->fPtr->sendingTries = 0;
    wfPtr->fPtr->endTime = wfPtr->fPtr->sendStartTime
                         = wfPtr->fPtr->sentTime = 0;
    wfPtr->records = 0;
    wfPtr->bufLen = 0;
    wfPtr->closing = 0;
    wfPtr->close = 0;

    /* set the timer to writeTimeout + some random # of secs <= 10 */
    skTimerCreate(&wfPtr->timerHandle,
                  writeTimeout,
                  timerHandler,
                  (void *) wfPtr);

    logMsg("Opened new file %s", wfPtr->fPtr->fName);

    return;
}


/*
 * popFreeQueue:
 *      Always from the head
 */
static fileStructPtr getFileStruct(void)
{
    fileStructPtr ptr;
    skDQErr_t err;
    uint8_t type;


    pthread_cleanup_push((cleanupHandler)pthread_mutex_unlock,
                         (void *)&spaceMutex);
    pthread_mutex_lock(&spaceMutex);
    do {
        if (unallocated_space[FC_RAM] >= allocFileSize) {
            type = FC_RAM;
            break;
        } else if (unallocated_space[FC_DISK] >= allocFileSize) {
            type = FC_DISK;
            break;
        }
        if (interfaceup) {
            shutdownInterface(NULL);
        }

        /* Wait for more space */
        pthread_cond_wait(&spaceCond, &spaceMutex);
    } while (1);

    if ((err = skDequePopFrontNB(freeQ, (void **)&ptr))) {
        if (err == SKDQ_EMPTY) {
            ptr = (fileStructPtr) calloc(sizeof(fileStruct), 1);
        } else {
            logMsg("Error: popFreeQueue");
            return NULL;
        }
    }
    ptr->isRamDisk = type;

    assert(unallocated_space[type] >= allocFileSize);
    unallocated_space[type] -= allocFileSize;

    if (!interfaceup) {
        restartInterface(NULL);
    }

    pthread_cleanup_pop(1);

    return ptr;
}


/*
 * addToFreeQueue:
 *      Add fPtr to the free queue.
 */
static void addToFreeQueue(fileStructPtr fPtr)
{
    pthread_cleanup_push((cleanupHandler)pthread_mutex_unlock,
                         (void *)&spaceMutex);
    pthread_mutex_lock(&spaceMutex);

    unallocated_space[fPtr->isRamDisk] += fPtr->size;
    skDequePushFront(freeQ, (void *)fPtr);

    /* Announce the availability of more space */
    pthread_cond_broadcast(&spaceCond);

    pthread_cleanup_pop(1);
}


/*
 * unloadFiles:
 *      Close all write files and flush write ram disk files to real disk.
 *      Then, flush all ram disk files in the send queue.  Since the
 *      file currently being sent is not in the send queue, this should
 *      be OK; i.e., we wont save and hence, double send the currently
 *      being sent file.
 */

static void unloadFiles(void)
{
    fileStructPtr fPtr;
    char srcFile[PATH_MAX], destFile[PATH_MAX];
    skDeque_t tmp;
    size_t i;


    for (i = 0; i < num_write_files; i++) {
        write_files[i].close = 1;
        closeWFile(&write_files[i], 0);
    }

    /* now check the send queues for ram disk files */
    tmp = skDequeCreate();
    while (!skDequePopFrontNB(highQ, (void **)&fPtr)) {
        if (fPtr->isRamDisk) {
            /* Copy to hard disk */
            sprintf(srcFile, "%s/%s", destDirs[FC_RAM], fPtr->fName);
            sprintf(destFile, "%s/%s", destDirs[FC_DISK], fPtr->fName);
            logMsg("Moving %s to %s.", srcFile, destFile);
            if (copyFile(srcFile, destFile) == 0) {
                DEBUGMSG("%s:%d Unlinking %s", __FILE__, __LINE__, srcFile);
                unlink(srcFile);
            } else {
                logMsg("Unable to copy %s/%s to %s",
                       destDirs[FC_RAM], fPtr->fName, destDirs[FC_DISK]);
            }
        }
        skDequePushBack(tmp, (void *)fPtr);
    }
    while (!skDequePopFrontNB(tmp, (void **)&fPtr)) {
        skDequePushBack(highQ, (void *)fPtr);
    }

    while (!skDequePopFrontNB(lowQ, (void **)&fPtr)) {
        if (fPtr->isRamDisk) {
            /* Copy to hard disk */
            sprintf(srcFile, "%s/%s", destDirs[FC_RAM], fPtr->fName);
            sprintf(destFile, "%s/%s", destDirs[FC_DISK], fPtr->fName);
            logMsg("Moving %s to %s.", srcFile, destFile);
            if (copyFile(srcFile, destFile) == 0) {
                DEBUGMSG("%s:%d Unlinking %s", __FILE__, __LINE__, srcFile);
                unlink(srcFile);
            } else {
                logMsg("Unable to copy %s/%s to %s",
                       destDirs[FC_RAM], fPtr->fName, destDirs[FC_DISK]);
            }
        }
        skDequePushBack(tmp, (void *)fPtr);
    }
    while (!skDequePopFrontNB(tmp, (void **)&fPtr)) {
        skDequePushBack(lowQ, (void *)fPtr);
    }
    skDequeDestroy(tmp);

    return;
}


void closeWFile(writeFileStructPtr wfPtr, uint8_t open)
{
    static pthread_mutex_t close_lock = PTHREAD_MUTEX_INITIALIZER;
    uint8_t quit = 0;

    /* Ah, the perils of threads.  wfPtr->closing keeps us from
       double-closing a wfPtr.  wfPtr->close makes sure we don't honor
       a request to close a wfPtr that has been closed and reopened
       since the request. */
    pthread_cleanup_push((cleanupHandler)pthread_mutex_unlock,
                         (void *)&close_lock);
    pthread_mutex_lock(&close_lock);

    if (wfPtr->closing || !wfPtr->close) {
        quit = 1;
    } else {
        wfPtr->closing = 1;
    }
    pthread_cleanup_pop(1);     /* unlock close_lock */

    if (quit) {
        logMsg("Avoiding duplicate call to closeWFile.");
        return;
    }

    LOGDEBUG(("Locking in closeWFile."));
    pthread_cleanup_push((cleanupHandler)pthread_mutex_unlock,
                         (void *)&wfPtr->mutex);
    pthread_mutex_lock(&wfPtr->mutex);
    LOGDEBUG(("Locked in closeWFile."));

    closeWFileBase(wfPtr, open);

    LOGDEBUG(("Unlocking in closeWFile."));
    pthread_cleanup_pop(1);     /* unlock wfPtr->mutex */
    LOGDEBUG(("Unlocked in closeWFile."));
}


/*
 * closeWFileBase:
 *      Close a writer file.
 */
static void closeWFileBase(writeFileStructPtr wfPtr, uint8_t open)
{
    off_t size;
    double change;

    logMsg("Closing %s", wfPtr->fPtr->fName);

    if(wfPtr->timerHandle) {
        skTimerDestroy(wfPtr->timerHandle);
        wfPtr->timerHandle = NULL;
    }

    if ((size = lzo_flush(wfPtr->lzofile)) == -1) {
        logMsg("Error flushing file.");
        logMsg(lzo_compr_strerror(wfPtr->lzofile));
        exit(EXIT_FAILURE);
    }

    wfPtr->fPtr->size += size;

    pthread_cleanup_push((cleanupHandler)pthread_mutex_unlock,
                         (void *)&spaceMutex);
    pthread_mutex_lock(&spaceMutex);

    unallocated_space[wfPtr->fPtr->isRamDisk] += allocFileSize - size;

    pthread_cond_broadcast(&spaceCond);

    pthread_cleanup_pop(1);

    wfPtr->fPtr->endTime = time((time_t *)NULL);

    if (wfPtr->fPtr->uncompressSize == 0) {
        change = 0.0;
    } else {
        change = (100.0 * ((double)wfPtr->fPtr->uncompressSize
                           - (double)wfPtr->fPtr->size)
                  / (double)wfPtr->fPtr->uncompressSize);
    }

    logMsg(("Closing file %s: %u seconds, %u records, "
            "%" PRIu64 " bytes, %4.1f%% compression"),
           wfPtr->fPtr->fName, wfPtr->fPtr->endTime - wfPtr->fPtr->startTime,
           wfPtr->records, wfPtr->fPtr->size, change);

    /* Close the file */
    if (fclose(wfPtr->gzFILE) == EOF) {
        logMsg("fclose error of %s/%s: %s", destDirs[wfPtr->fPtr->isRamDisk],
               wfPtr->fPtr->fName, strerror(errno));
        /* What do we do now? */
    }

    wfPtr->gzFILE = NULL;

    /* In the next line, the sizeof(uint32_t) is there to account for
       the end of compression marker. */
    if (wfPtr->fPtr->size <= wfPtr->headerSize + sizeof(uint32_t)) {
        char buffer[PATH_MAX];

        if (snprintf(buffer, sizeof(buffer), "%s/%s",
                     destDirs[wfPtr->fPtr->isRamDisk],
                     wfPtr->fPtr->fName) >= (int)sizeof(buffer)) {
            logMsg("Filename exceeded buffer size.");
            exit(EXIT_FAILURE);
        }
        logMsg("Removing empty file %s", wfPtr->fPtr->fName);
        DEBUGMSG("%s:%d Unlinking %s", __FILE__, __LINE__, buffer);
        unlink(buffer);
        addToFreeQueue(wfPtr->fPtr);
    } else if (skProbeGetPriority(wfPtr->probe) > 50) {
        addToHighPriorityQueue(wfPtr->fPtr);
    } else {
        addToLowPriorityQueue(wfPtr->fPtr);
    }

    logMsg("Finished closing %s", wfPtr->fPtr->fName);

    if (open) {
        openWFileBase(wfPtr);
    }


    return;
}


/*
 * startReader:
 *      Setup for reading from the netflow IP +  UDP port socket.
 * Inputs: None
 * Outputs:  1 if failure. 0 if ok
 */
static int startReader(void)
{
    size_t i;
    pduSourcePool_t pool;

    pool = pduSourcePoolCreate();
    if (pool == NULL) {
        return 1;
    }

    reading = 1;

    for (i = 0 ; i < num_write_files; i++) {

        switch (skProbeGetType(write_files[i].probe)) {
          case PROBE_ENUM_NETFLOW:
            {
                write_files[i].source.pdu =
                    pduSourceCreateFromProbeDef(pool,
                                                write_files[i].probe,
                                                32768,
                                                8388608 / num_write_files,
                                                logMsg);
                if (write_files[i].source.pdu == NULL)
                {
                    return 1;
                }
            }
            break;
#ifdef HAVE_YAF_YAFCORE_H
          case PROBE_ENUM_YAF:
            {
                in_addr_t addr;
                uint16_t port;
                int rv;

                rv = skProbeGetListenAsHost(&addr, &port,
                                            write_files[i].probe);
                if (rv != 0) {
                    return 1;
                }

                write_files[i].source.yaf =
                    yafSourceCreate(port,
                                    addr,
                                    8388608 / num_write_files,
                                    logMsg);
                
                if (write_files[i].source.yaf == NULL)
                {
                    return 1;
                }
            }
            break;
#endif
          default:
            logMsg("Unsupported probe type %d",
                   (int)skProbeGetType(write_files[i].probe));
            assert(0);
            abort();
        } /* switch () */

        write_files[i].valid_source = 1;

        if (pthread_create(&write_files[i].reader_thread,
                           NULL,
                           reader,
                           (void *)&write_files[i]))
        {
            return 1;
        }
        write_files[i].running = 1;
    } /* for (i) */

    pduSourcePoolDestroy(pool);

    return 0;
}

static void fileStructPtr_name(void *item, char *path, char *name)
{
    size_t size;

    fileStructPtr fPtr = (fileStructPtr)item;

    curSendFPtr = fPtr;

    size = snprintf(path, PATH_MAX, "%s/%s", destDirs[fPtr->isRamDisk],
                    fPtr->fName);
    if (size >= PATH_MAX) {
        logMsg("Path too long.  Exiting.");
        exit(EXIT_FAILURE);
    }
    name[0] = '\0';
    fPtr->sendStartTime = time((time_t *)NULL);
}


static void sender_callback(
    char           *path,
    void           *item,
    sktErr_t        err,
    skTransfer_t    UNUSED(sender))
{
    fileStructPtr fPtr = (fileStructPtr)item;
    char *retry = "";

    switch (err) {
      case SK_TRANSFER_ESUCCESS:
        logMsg("Successfully sent %s", path);
        fPtr->sentTime = time((time_t *)NULL);
        dumpStats(fPtr);
        curSendFPtr = NULL;
        addToFreeQueue(fPtr);
        DEBUGMSG("%s:%d Unlinking %s", __FILE__, __LINE__, path);
        unlink(path);
        break;
      case SK_TRANSFER_ETIMEOUT:
      case SK_TRANSFER_EFAILED:
        fPtr->sendingTries++;
        if (fPtr->sendingTries < 3) {
            retry = " (Retrying)";
            addToSendQueue(fPtr);
        }
        logMsg("Failed to transfer %s.%s", path, retry);
        break;
      case SK_TRANSFER_EFATAL:
        logMsg("Fatally failed to transfer %s.", path);
        curSendFPtr = NULL;
        addToFreeQueue(fPtr);
        DEBUGMSG("%s:%d Unlinking %s", __FILE__, __LINE__, path);
        unlink(path);
        break;
      case SK_TRANSFER_ESYSTEM:
      case SK_TRANSFER_ESYSTEM_THREAD:
        logMsg("Fatal system error in file transfer.  Shutting down.");
        exit(EXIT_FAILURE);
      case SK_TRANSFER_ESHUTDOWN_EXTERNAL:
        logMsg("File transfer aborted.");
        break;
      default:
        abort();                  /* Should never get here */
    }
}


/* return 1 if failure. 0 if ok */
static int startSender(void)
{
    if (skCreateSenderServer(&sender, fc_listen_port)
        != SK_TRANSFER_ESUCCESS)
    {
        return 1;
    }
    if (fc_client_addr == INADDR_ANY) {
        if (skServerValidClientAny(sender) != SK_TRANSFER_ESUCCESS) {
            goto error;
        }
    } else if (skServerValidClientAdd(
                   sender, fc_client_addr) != SK_TRANSFER_ESUCCESS)
    {
        goto error;
    }
    if (skTransferSetQueue(sender, sendQ) != SK_TRANSFER_ESUCCESS) {
        goto error;
    }
    if (skTransferSetAckTimeout(sender, ackTimeout) != SK_TRANSFER_ESUCCESS) {
        goto error;
    }
    if (skTransferSetCallback(sender, sender_callback) != SK_TRANSFER_ESUCCESS)
    {
        goto error;
    }
    if (skTransferSetLogFn(sender, logMsg) != SK_TRANSFER_ESUCCESS) {
        goto error;
    }
    if (skSenderSetFilenameFn(
            sender, fileStructPtr_name) != SK_TRANSFER_ESUCCESS)
    {
        goto error;
    }
    if (skTransferStart(sender) != SK_TRANSFER_ESUCCESS) {
        goto error;
    }

    return 0;
  error:
    skDestroyTransfer(sender);
    return 1;
}


static void stopSender(void)
{
    if (sender != NULL) {
        logMsg("Stopping sender.");
        skTransferStop(sender, 10);
        skDestroyTransfer(sender);
    }
    sender = NULL;
    return;
}

/* Stop the reader */
static void stopReader(void)
{
    size_t i;

    if (!reading) {
        return;
    }

    logMsg("Stopping readers. highQ.len = %d  lowQ.len = %d",
           skDequeSize(highQ), skDequeSize(lowQ));

    reading = 0;

    for (i = 0; i < num_write_files; i++) {

        if (write_files[i].valid_source) {
            switch (skProbeGetType(write_files[i].probe)) {
              case PROBE_ENUM_NETFLOW:
                pduSourceDestroy(write_files[i].source.pdu);
                break;
#ifdef HAVE_YAF_YAFCORE_H
              case PROBE_ENUM_YAF:
                yafSourceDestroy(write_files[i].source.yaf);
                break;
#endif
              default:
                assert(0);
            }
            write_files[i].valid_source = 0;
        }

        if (write_files[i].running) {
            pthread_join(write_files[i].reader_thread, NULL);
            write_files[i].running = 0;
        }
    }

    return;
}

static void dumpStats(fileStructPtr fPtr)
{
    logMsg("%s Write %u secs. Wait %u secs. Send %u secs.",
           fPtr->fName, (fPtr->endTime - fPtr->startTime),
           (fPtr->sendStartTime - fPtr->endTime),
           (fPtr->sentTime - fPtr->sendStartTime));
    return;
}


static void shutdownInterface(void *UNUSED(dummy))
{
    logMsg("Shutting down netflow interface.");
    stopReader();
    interfaceup = 0;
    return;
}


static void restartInterface(void *UNUSED(dummy))
{
    logMsg("Restarting netflow interface.");
    if (startReader()) {
        logMsg("Couldn't restart reader.");
        exit(EXIT_FAILURE);
    }
    interfaceup = 1;
    return;
}


static void initQueues(void)
{
    lowQ = skDequeCreate();
    highQ = skDequeCreate();
    freeQ = skDequeCreate();
    sendQ = skDequeCreateMerged(lowQ, highQ);
}


static void destroyQueues(void)
{
    skDequeDestroy(sendQ);
    skDequeDestroy(freeQ);
    skDequeDestroy(highQ);
    skDequeDestroy(lowQ);
}


/* Program entry point. */
int main(int argc, char **argv)
{
    appSetup(argc, argv);               /* never returns on failure */

    sklogOpen();

    /*  Initialize the queues */
    initQueues();

    /* start the logger and become a daemon */
#ifdef DEBUG
    skdaemonDontFork();
#endif
    if (skdaemonize(&shuttingDown, &appTeardown) == -1) {
        exit(EXIT_FAILURE);
    }

    /* Populate the queues */
    loadFiles();

    /* Create the write files */
    createWriteFiles();

    /* Store the main thread ID */
    main_thread = pthread_self();

    /* Start the sender thread */
    if (startSender()) {
        logMsg("Failed to start sender. Exiting.");
        skAppPrintErr("Failed to start sender. Exiting.");
        exit(EXIT_FAILURE);
    }

    /* Start the reader threads */
    if (startReader()) {
        logMsg("Failed to start reader. Exiting.");
        skAppPrintErr("Failed to start reader. Exiting.");
        exit(EXIT_FAILURE);
    }

    /* We now run forever, excepting signals */
    while (!shuttingDown) {
        pause();
    }

    appTeardown();

    exit(EXIT_SUCCESS);         /* Make sure we exit even if we have
                                   to tear down threads. */
    return 0;
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
