/*
** Copyright (C) 2004-2006 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Add your description here
**
*/

#include "silk.h"

RCSIDENT("$SiLK: reader.c 5217 2006-10-04 17:00:35Z mwd $");

#include "fcinternals.h"
#include "sklog.h"

/*
** reader:
**      Reads records from a wfPtr source.
**      Packs and write all real records according to type.
**      If full (i.e., past the high water mark), close file and put
**      into send queue at tail.
**      Open a new file.
*/
void *reader(void *vwfPtr) {
    sigset_t sigs;
    writeFileStructPtr wfPtr = (writeFileStructPtr)vwfPtr;
    volatile pdu_source_breakable_t rv = PDUS_BREAKABLE;
    probe_enum_t type;
    volatile ssize_t rec_size;

    assert(wfPtr);

    /* Turn off all signals except for SIGPIPE. */
    sigfillset(&sigs);
    sigdelset(&sigs, SIGPIPE);
    pthread_sigmask(SIG_SETMASK, &sigs, NULL); /* Only handle SIGPIPE */

    logMsg("%s packer for %s started.",
           skProbeGetName(wfPtr->probe), skProbeGetSensorName(wfPtr->probe));

    type = skProbeGetType(wfPtr->probe);

    rec_size = FC_REC_SIZE(wfPtr->version);

    /* Infloop */
    while (reading) {
        uint8_t rec[FC_MAX_SIZE];
        flowcapRec_t *fcr = (flowcapRec_t *)rec;

        /* Get the next flowcap record */
        switch (type) {

          case PROBE_ENUM_NETFLOW:
            rv = pduSourceGetFlowcap(wfPtr->source.pdu, fcr, wfPtr->version);
            if (rv == PDUS_ABORT) {
                goto endloop;
            }
            break;


#ifdef HAVE_YAF_YAFCORE_H
          case PROBE_ENUM_YAF:
            {
                int iv = yafSourceGetFlowcap(wfPtr->source.yaf, 
                                             fcr, wfPtr->version);
                if (iv == -1) {
                    goto endloop;
                }
            }
            break;
#endif

          default:
            assert(0);
            abort();
        }

        pthread_cleanup_push((cleanupHandler)pthread_mutex_unlock,
                             (void *)&wfPtr->mutex);
        pthread_mutex_lock(&wfPtr->mutex);

        /* Write the record to the file */
        if (lzo_write(wfPtr->lzofile, rec, rec_size) != rec_size) {
            logMsg("Failed compressed write.");
            logMsg(lzo_compr_strerror(wfPtr->lzofile));
            exit(EXIT_FAILURE);
        }
        wfPtr->fPtr->uncompressSize += rec_size;
        wfPtr->records++;

        /* Check to see if we have reached the size limit */
        if ((rv == PDUS_BREAKABLE) &&
            (lzo_compr_size_upper_bound(wfPtr->lzofile) >= wfPtr->HWM)) {
            wfPtr->close = 1;
        }

        pthread_cleanup_pop(1);       /* unlock wfPtr->mutex */

        /* Close and acquire new file if necessary */
        if (wfPtr->close) {
            /* Close and ship the file */
            closeWFile(wfPtr, 1);
        }
    }

  endloop:

    logMsg("%s packer for %s ended.",
           skProbeGetName(wfPtr->probe), skProbeGetSensorName(wfPtr->probe));

    /* End thread */
    return NULL;
}

/*
** dumpHeader:
**      Write the header to the file.   All info  required is in wfPtr
*/
uint32_t dumpHeader(writeFileStructPtr wfPtr)
{
    fcHeader_V2_t hdr;

    assert(FC_REC_SIZE(wfPtr->version) != -1);

    memset(&hdr, 0, sizeof(hdr));

    PREPHEADER(&hdr.gHdr);
    hdr.gHdr.isBigEndian = 1;
    hdr.gHdr.version = wfPtr->version;
    hdr.gHdr.type = wfPtr->type;
    hdr.gHdr.compMethod = SK_COMPMETHOD_LZO1X;

    strcpy(hdr.sensorID, skProbeGetSensorName(wfPtr->probe));
    strcpy(hdr.probeID, skProbeGetName(wfPtr->probe));

    fwrite(&hdr, sizeof(hdr), 1, wfPtr->gzFILE);

    return sizeof(hdr);
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
