/*
** Copyright (C) 2006 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Conversion functions for YAF
**
*/


#include "silk.h"

RCSIDENT("$SiLK: convert-yaf.c 4876 2006-09-14 21:18:03Z mthomas $");

#include "convert-yaf.h"
#include "convert-flowcap.h"

int yafToFlowcap(
    const YAFlow   *yafPtr,
    yaf_dir_t       direction,
    flowcapRec_t   *flowcap_rec,
    int             flowcap_version)
{
    rwRec rwrec;
    int rv;

    /* I am tired of making these functions.  As such, I am paving the
       way towards making flowcap records something librw can handle.
       We will use an rwRec as an intermediate. */

    rv = yafToGeneric(yafPtr, direction, &rwrec);
    if (rv) {
        return rv;
    }
    return genericToFlowcap(&rwrec, flowcap_rec, flowcap_version);
}


int yafToGeneric(
    const YAFlow   *yaf,
    yaf_dir_t       direction,
    rwRec          *rwrec)
{
    const YAFlowVal *fv;
    uint64_t stime, etime;
    imaxdiv_t sdiv, ediv;

    assert(yaf);
    assert(rwrec);

    switch (direction) {
      case YAF_FORWARD:
        fv = &yaf->val;
        break;
      case YAF_REVERSE:
        fv = &yaf->rval;
        break;
    }

    if (fv->pkt == 0) {
        assert(direction != YAF_FORWARD);
        return -1;
    }

    memset(rwrec, 0, sizeof(*rwrec));

    /* sip, dip, sport, dport */
    switch (direction) {
      case YAF_FORWARD:
        rwrec->sIP.ipnum = yaf->key.sip;
        rwrec->dIP.ipnum = yaf->key.dip;
        rwrec->sPort     = yaf->key.sp;
        rwrec->dPort     = yaf->key.dp;
        break;
      case YAF_REVERSE:
        rwrec->sIP.ipnum = yaf->key.dip;
        rwrec->dIP.ipnum = yaf->key.sip;
        rwrec->sPort     = yaf->key.dp;
        rwrec->dPort     = yaf->key.sp;
        break;
    }
    /* proto */
    rwrec->proto = yaf->key.proto;

    /* output gets 802.1Q VLAN tag, for now */
    rwrec->output = fv->tag;

    assert(sizeof(sdiv.quot) >= sizeof(yaf->stime));
    assert(sizeof(ediv.quot) >= sizeof(yaf->etime));

    /* time calculations */
    stime = yaf->stime;
    if (direction == YAF_REVERSE) {
        stime += yaf->rdtime;
    }
    etime = yaf->etime - stime;

    sdiv = imaxdiv(stime, 1000);
    ediv = imaxdiv(etime, 1000);

    /* stime */
    if (sdiv.quot > UINT32_MAX) {
        rwrec->sTime      = UINT32_MAX;
        rwrec->sTime_msec = 999;
    } else {
        rwrec->sTime      = sdiv.quot;
        rwrec->sTime_msec = sdiv.rem;
    }

    /* etime */
    if (ediv.quot > UINT32_MAX) {
        rwrec->elapsed      = UINT32_MAX;
        rwrec->elapsed_msec = 999;
    } else {
        rwrec->elapsed      = ediv.quot;
        rwrec->elapsed_msec = ediv.rem;
    }

    /* pkts */
    if (fv->pkt > UINT32_MAX) {
        rwrec->pkts = UINT32_MAX;
    } else {
        rwrec->pkts = fv->pkt;
    }

    /* bytes */
    if (fv->oct > UINT32_MAX) {
        rwrec->bytes = UINT32_MAX;
    } else {
        rwrec->bytes = fv->oct;
    }

    /* flags */
    rwrec->flags      = fv->iflags | fv->uflags;
    rwrec->init_flags = fv->iflags;
    rwrec->rest_flags = fv->uflags;
    rwrec->tcp_state |= SK_TCPSTATE_EXPANDED;

    if ((yaf->reason & YAF_END_MASK) == YAF_END_ACTIVE) {
        rwrec->tcp_state |= SK_TCPSTATE_TIMEOUT_KILLED;
    }
    if (yaf->reason & YAF_ENDF_ISCONT) {
        rwrec->tcp_state |= SK_TCPSTATE_TIMEOUT_STARTED;
    }

    return 0;
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
