/*
** Copyright (C) 2006 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Interface to pull a single flow from YAF
**
*/


#include "silk.h"

RCSIDENT("$SiLK: yafsource.c 5178 2006-10-03 15:47:12Z mwd $");

#include "utils.h"
#include "circbuf.h"
#include "yafsource.h"
#include "convert-yaf.h"

struct yaf_source_t {
    flowcap_stats_t statistics;
    pthread_mutex_t stats_mutex;

    yaf_dir_t dir;
    YAFlow *yaf;

    in_addr_t listen_addr;
    int port;

    int pipe[2];

    FixReader *reader;

    pthread_t thread;
    uint32_t bufsize;
    circBuf_t data_buffer;
    void *tmp_buffer;
    sk_msg_fn_t logfn;
    pthread_mutex_t mutex;
    pthread_cond_t cond;
    unsigned destroyed : 1;
};


/* Mutex around calls to yaf_read_start_tcp. */
static pthread_mutex_t yaf_start_mutex = PTHREAD_MUTEX_INITIALIZER;


static void *yaf_reader(void *vsource)
{
    yafSource_t source = (yafSource_t)vsource;
    struct sockaddr_in addr;
    int sock;
    int rv;
    int maxfd;
    fd_set set;
    int fd = -1;

    pthread_mutex_lock(&source->mutex);

    sock = socket(AF_INET, SOCK_STREAM, 0);

    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(source->listen_addr);
    addr.sin_port = htons(source->port);

    rv = bind(sock, (struct sockaddr *)&addr, sizeof(addr));
    if (rv == -1 && source->logfn) {
        source->logfn("yaf_reader: bind failed (%s)", strerror(errno));
    }
    if (rv == 0) {
        rv = listen(sock, 1);
        if (rv == -1 && source->logfn) {
            source->logfn("yaf_reader: listen failed (%s)", strerror(errno));
        }
    }
    if (rv == -1) {
        close(sock);
        source->destroyed = 1;
        pthread_cond_signal(&source->cond);
        pthread_mutex_unlock(&source->mutex);
        return NULL;
    }

    FD_ZERO(&set);
    FD_SET(sock, &set);
    FD_SET(source->pipe[0], &set);
    maxfd = ((sock > source->pipe[0]) ? sock : source->pipe[0]) + 1;

    source->tmp_buffer = (void *)circBufNextHead(source->data_buffer);
    assert(source->tmp_buffer != NULL);

    pthread_cond_signal(&source->cond);
    pthread_mutex_unlock(&source->mutex);

    while (!source->destroyed) {
        fd_set rset = set;

        fd = select(maxfd, &rset, NULL, NULL, NULL);
        if (fd == -1) {
            if (errno == EINTR) {
                continue;
            }
            source->destroyed = 1;
            break;
        }
        if (FD_ISSET(source->pipe[0], &rset)) {
            uint8_t c;

            pthread_mutex_lock(&source->mutex);
            rv = read(source->pipe[0], &c, 1);
            source->destroyed = 1;
            pthread_mutex_unlock(&source->mutex);
            break;
        }
        if (FD_ISSET(sock, &rset)) {
            struct sockaddr_in addr;
            GError *err = NULL;
            socklen_t len = sizeof(addr);
            char ip[SK_NUM2DOT_STRLEN];

            fd = accept(sock, (struct sockaddr *)&addr, &len);
            if (fd == -1) {
                continue;
            }

            num2dot_r(ntohl(addr.sin_addr.s_addr), ip);

            if (source->logfn) {
                source->logfn("YAF source connected: %s", ip);
            }

            pthread_mutex_lock(&source->mutex);

            /* yaf_read_start_tcp creates a static data structure the
               first time it is called.  Mutexed to prevent it from
               being created twice.  (Unlinkely, but possible.) */
            pthread_mutex_lock(&yaf_start_mutex);
            source->reader = yaf_read_start_tcp(fd, &err);
            pthread_mutex_unlock(&yaf_start_mutex);

            if (source->reader == NULL) {
                if (err && source->logfn) {
                    source->logfn("yaf_read_start_tcp: %s", err->message);
                }
                g_clear_error(&err);
            } else if (source->reader != NULL) {

                pthread_mutex_unlock(&source->mutex);

                while (!source->destroyed) {
                    gboolean rv;
                    GError *err = NULL;

                    rv = yaf_read(source->reader, (YAFlow *)source->tmp_buffer,
                                  &err);

                    if (!rv) {
                        if (err &&
                            err->domain == FIX_ERROR_DOMAIN &&
                            err->code == FIX_ERROR_NLREAD)
                        {
                            g_clear_error(&err);
                            continue;
                        }
                        if (err && source->logfn) {
                            source->logfn("yaf_read: %s", err->message);
                        }
                        g_clear_error(&err);
                        break;
                    }

                    /* Advance the circular buffer */
                    source->tmp_buffer =
                        (void *)circBufNextHead(source->data_buffer);
                    if (source->tmp_buffer == NULL) {
                        assert(source->destroyed);
                        break;
                    }
                }

                pthread_mutex_lock(&source->mutex);

                if (source->reader) {
                    fix_reader_free(source->reader);
                    source->reader = NULL;
                }

            }

            pthread_mutex_unlock(&source->mutex);

            if (source->logfn) {
                source->logfn("YAF source disconnected: %s", ip);
            }

            fd = -1;
        }
    }
    if (fd != -1) {
        close (fd);
    }
    close (sock);

    pthread_mutex_lock(&source->mutex);
    while (!source->destroyed) {
        pthread_cond_wait(&source->cond, &source->mutex);
    }
    pthread_cond_signal(&source->cond);
    pthread_mutex_unlock(&source->mutex);

    return NULL;
}


yafSource_t yafSourceCreate(
    int port,
    in_addr_t listen_on_address,
    uint32_t max_flows,
    sk_msg_fn_t logfn)
{
    yafSource_t source;
    int rv;

    source = (yafSource_t)calloc(1, sizeof(*source));
    if (source == NULL) {
        return NULL;
    }
    source->logfn = logfn;
    source->listen_addr = listen_on_address;
    source->port = port;

    pthread_mutex_init(&source->stats_mutex, NULL);
    pthread_mutex_init(&source->mutex, NULL);
    pthread_cond_init(&source->cond, NULL);

    source->data_buffer = circBufCreate(sizeof(*source->yaf), max_flows);
    if (source->data_buffer == NULL) {
        free(source);
        return NULL;
    }

    source->bufsize = max_flows;
    rv = pipe(source->pipe);
    if (rv != 0) {
        circBufDestroy(source->data_buffer);
        free(source);
        return NULL;
    }

    pthread_mutex_lock(&source->mutex);
    rv = pthread_create(&source->thread, NULL, yaf_reader, (void *)source);
    if (rv != 0) {
        circBufDestroy(source->data_buffer);
        close(source->pipe[0]);
        close(source->pipe[1]);
        free(source);
        return NULL;
    }
    pthread_cond_wait(&source->cond, &source->mutex);
    if (source->destroyed) {
        circBufDestroy(source->data_buffer);
        pthread_mutex_unlock(&source->mutex);
        pthread_join(source->thread, NULL);
        close(source->pipe[0]);
        close(source->pipe[1]);
        free(source);
        return NULL;
    }
    pthread_mutex_unlock(&source->mutex);

    return source;
}


void yafSourceDestroy(
    yafSource_t source)
{
    static uint8_t c = 0;

    assert(source);

    pthread_mutex_lock(&source->mutex);
    source->destroyed = 1;
    circBufDestroy(source->data_buffer);

    write(source->pipe[1], &c, 1);

    if (source->reader) {
        fix_reader_free(source->reader);
        source->reader = NULL;
    }

    pthread_cond_broadcast(&source->cond);
    pthread_cond_wait(&source->cond, &source->mutex);
    pthread_join(source->thread, NULL);
    pthread_mutex_unlock(&source->mutex);

    pthread_mutex_destroy(&source->mutex);
    pthread_cond_destroy(&source->cond);
    close(source->pipe[0]);
    close(source->pipe[1]);
    free(source);
}


int yafSourceGetGeneric(
    yafSource_t     source,
    rwRec          *rwrec)
{
    int rv;
    assert(source);

  again:
    if (source->yaf && source->dir == YAF_FORWARD) {
        source->dir = YAF_REVERSE;
    } else {
        source->dir = YAF_FORWARD;
        source->yaf = (YAFlow *)circBufNextTail(source->data_buffer);
        if (source->yaf == NULL) {
            return -1;
        }
    }

    rv = yafToGeneric(source->yaf, source->dir, rwrec);
    if (rv == -1) {
        if (source->dir == YAF_FORWARD) {
            return -1;
        }
        goto again;
    }

    return 0;
}

int yafSourceGetFlowcap(
    yafSource_t     source,
    flowcapRec_t   *fc_rec,
    int             fc_version)
{
    int rv;
    assert(source);

  again:
    if (source->yaf && source->dir == YAF_FORWARD) {
        source->dir = YAF_REVERSE;
    } else {
        source->dir = YAF_FORWARD;
        source->yaf = (YAFlow *)circBufNextTail(source->data_buffer);
        if (source->yaf == NULL) {
            return -1;
        }
    }

    rv = yafToFlowcap(source->yaf, source->dir, fc_rec, fc_version);
    if (rv == -1) {
        if (source->dir == YAF_FORWARD) {
            return -1;
        }
        goto again;
    }

    return 0;
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
