/*
** Copyright (C) 2001-2006 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  mapsid.c
**    dump a map of sid integers to names from silk_site.h
*/

#include "silk.h"

RCSIDENT("$SiLK: mapsid.c 4301 2006-07-19 21:50:10Z mthomas $");

#include "utils.h"
#include "sksite.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to write --help output */
#define USAGE_FH stdout

/* where to send output */
#define OUT_FH stdout


/* LOCAL FUNCTIONS */

static void appUsageLong(void);


/* OPTIONS SETUP */

/*
** typedef enum {
** } appOptionsEnum;
*/

static struct option appOptions[] = {
    {0,0,0,0}           /* sentinel entry */
};

static const char *appHelp[] = {
    (char *)NULL
};


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  This function is
 *    passed to optionsSetup(); optionsParse() will call this funciton
 *    and then exit the program when the --help option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                       \
    ("[SENSORS]\n"                                                      \
     "\tMaps between sensor names and sensor IDs.  Prints a list of\n"  \
     "\tall sensors when no command line arguments are given.\n")

    skAppStandardUsage(USAGE_FH, USAGE_MSG, appOptions, appHelp);
}


static void printByNameOrNumber(char *sensor)
{
    int err;
    uint32_t temp;
    sensorID_t sid;
    char s_name[SK_MAX_STRLEN_SENSOR+1];

    /* try to parse as a number */
    err = skStringParseUint32(&temp, sensor, 0, SK_INVALID_SENSOR-1);
    if (err == -1 || err == -2) {
        skAppPrintErr("No name or number was provided.");
    } else if (err == -4) {
        skAppPrintErr(("number '%s' is not a valid sensor number:"
                       " too large to parse"),
                      sensor);
    } else if (err == -12) {
        skAppPrintErr("number '%s' is not a valid sensor number: out of range",
                      sensor);
    } else if (err == 0) {
        /* got a clean parse */
        sid = (sensorID_t)temp;
        if (sksiteSensorGetName(s_name, sid, sizeof(s_name)) == NULL) {
            skAppPrintErr("number '%s' is not a valid sensor number",
                          sensor);
        } else {
            fprintf(OUT_FH, "%5u -> %s\n",
                    sid, s_name);
        }
    } else {
        /* didn't get a clean parse. try to treat as a name */
        sid = sksiteLookupSensor(sensor);
        if (sid == SK_INVALID_SENSOR) {
            /* for legacy reasons, up-case the sensor and try again */
            upper(sensor);
            sid = sksiteLookupSensor(sensor);
        }

        if (sid == SK_INVALID_SENSOR) {
            skAppPrintErr("name '%s' is not a valid sensor name", sensor);
        } else {
            fprintf(OUT_FH, "%s -> %5u\n", sensor, sid);
        }
    }
}


static void printAllSensors(void)
{
    sensorID_t sid = 0;
    char s_name[SK_MAX_STRLEN_SENSOR+1];

    while (sksiteSensorGetName(s_name, sid, sizeof(s_name))) {
        fprintf(OUT_FH, "%5u -> %s\n", sid, s_name);
        ++sid;
    }

    fprintf(OUT_FH, "Total sensors %u\n", sid);
}


int main(int argc, char **argv)
{
    int i;

    skAppRegister(argv[0]);

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* Set up the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                     /* never returns */
    }

    /* register the options and the handler */
    /*
    ** if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL))
    ** {
    **     skAppPrintErr("unable to register application options");
    **     exit(EXIT_FAILURE);
    ** }
    */

    /* parse the options---we have none of our own, but this will
     * check for the --help and --version options */
    i = optionsParse(argc, argv);
    if (i < 0) {
        /* options parsing should print error */
        skAppUsage();           /* never returns */
    }

    /* done with options */
    optionsTeardown();

    if (argc > i) {
        for ( ; i < argc; i++) {
            printByNameOrNumber(argv[i]);
        }
    } else {
        /* no args. dump all */
        printAllSensors();
    }

    return 0;
}
