/*
** Copyright (C) 2004-2006 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwbag
**
**    Build binary Bag files from flow records.
**
**    For IPs, use the standard Bag API to create the bags.  For Bags
**    of ports or protocols, store the data in an in-core array, then
**    convert it to a bag when the time comes to store it on disk.
**
*/


#include "silk.h"

RCSIDENT("$SiLK: rwbag.c 5910 2006-12-11 19:44:49Z mthomas $");

#include "utils.h"
#include "rwpack.h"
#include "sksite.h"
#include "bagtree.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to write --help output */
#define USAGE_FH stdout

/* number of potential IP bags: {sIP,dIP} * {flows,pkts,bytes} */
#define NUM_OUTPUTS 6

/* number of potential arrays: {sPort,dPort,proto} * {flows,pkts,bytes} */
#define NUM_ARRAY_OUTPUTS 9

typedef struct {
    skBag_header_t     *bag;
    skstream_t         *stream;
} bagfile_t;

typedef struct {
    skBag_counter_t    *array;
    skBag_key_t         size;
    skstream_t         *stream;
} arrayfile_t;

/* EXPORTED FUNCTIONS */


/* LOCAL FUNCTIONS */

static void appUsageLong(void);              /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv); /* never returns when error */
static int  appOptionsHandler(clientData, int, const char *);
static int  prepareBagFile(bagfile_t*, const char *);
static int  prepareArrayBagFile(arrayfile_t*, const char *);


/* LOCAL VARIABLES */

/* The potential IP bag files to create */
static bagfile_t bag_io[NUM_OUTPUTS];

/* The potential portbag files to create */
static arrayfile_t array_io[NUM_ARRAY_OUTPUTS];

/* the compression method to use when writing the files.
 * sksiteCompmethodOptionsRegister() will set this to the default or
 * to the value the user specifies. */
static sk_compmethod_t comp_method;

/* First file index in argv[] after handling all options */
static int  first_file;

/*
 * stdout_used is set to 1 by prepareBagFile and prepareArrayBagFile
 * when a file is to be written to stdout.
 */
static int stdout_used = 0;

/* True if an overflow warning has been printed for given stream */
static int overflow_warning[NUM_OUTPUTS];
static int array_overflow_warning[NUM_ARRAY_OUTPUTS];


/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    /* These MUST be kept in order with the options */
    SIP_FLOWS=0,  SIP_PKTS,  SIP_BYTES,
    DIP_FLOWS, DIP_PKTS, DIP_BYTES,
    SPORT_FLOWS, SPORT_PKTS, SPORT_BYTES,
    DPORT_FLOWS, DPORT_PKTS, DPORT_BYTES,
    PROTO_FLOWS, PROTO_PKTS, PROTO_BYTES
} appOptionsEnum;

/* This #define is used to adjust the indices for the array-based
 * bags.  It must be the first enum item that represents an array bag
 * instead of a tree bag. */
#define AB_OFFSET SPORT_FLOWS

static struct option appOptions[] = {
    {"sip-flows",       REQUIRED_ARG, 0, SIP_FLOWS},
    {"sip-packets",     REQUIRED_ARG, 0, SIP_PKTS},
    {"sip-bytes",       REQUIRED_ARG, 0, SIP_BYTES},
    {"dip-flows",       REQUIRED_ARG, 0, DIP_FLOWS},
    {"dip-packets",     REQUIRED_ARG, 0, DIP_PKTS},
    {"dip-bytes",       REQUIRED_ARG, 0, DIP_BYTES},
    {"sport-flows",     REQUIRED_ARG, 0, SPORT_FLOWS},
    {"sport-packets",   REQUIRED_ARG, 0, SPORT_PKTS},
    {"sport-bytes",     REQUIRED_ARG, 0, SPORT_BYTES},
    {"dport-flows",     REQUIRED_ARG, 0, DPORT_FLOWS},
    {"dport-packets",   REQUIRED_ARG, 0, DPORT_PKTS},
    {"dport-bytes",     REQUIRED_ARG, 0, DPORT_BYTES},
    {"proto-flows",     REQUIRED_ARG, 0, PROTO_FLOWS},
    {"proto-packets",   REQUIRED_ARG, 0, PROTO_PKTS},
    {"proto-bytes",     REQUIRED_ARG, 0, PROTO_BYTES},
    {0,0,0,0}           /* sentinel entry */
};

/* allow these older options */
static struct option oldOptions[] = {
    {"sf-file",         REQUIRED_ARG, 0, SIP_FLOWS},
    {"sp-file",         REQUIRED_ARG, 0, SIP_PKTS},
    {"sb-file",         REQUIRED_ARG, 0, SIP_BYTES},
    {"df-file",         REQUIRED_ARG, 0, DIP_FLOWS},
    {"dp-file",         REQUIRED_ARG, 0, DIP_PKTS},
    {"db-file",         REQUIRED_ARG, 0, DIP_BYTES},
    {"port-sf-file",    REQUIRED_ARG, 0, SPORT_FLOWS},
    {"port-sp-file",    REQUIRED_ARG, 0, SPORT_PKTS},
    {"port-sb-file",    REQUIRED_ARG, 0, SPORT_BYTES},
    {"port-df-file",    REQUIRED_ARG, 0, DPORT_FLOWS},
    {"port-dp-file",    REQUIRED_ARG, 0, DPORT_PKTS},
    {"port-db-file",    REQUIRED_ARG, 0, DPORT_BYTES},
    {"proto-f-file",    REQUIRED_ARG, 0, PROTO_FLOWS},
    {"proto-p-file",    REQUIRED_ARG, 0, PROTO_PKTS},
    {"proto-b-file",    REQUIRED_ARG, 0, PROTO_BYTES},
    /* allow proto-X to work as a shortcut since it is unique*/
    {"proto-f",         REQUIRED_ARG, 0, PROTO_FLOWS},
    {"proto-p",         REQUIRED_ARG, 0, PROTO_PKTS},
    {"proto-b",         REQUIRED_ARG, 0, PROTO_BYTES},
    {0,0,0,0}           /* sentinel */
};

static const char *appHelp[] = {
    "Write bag of flow counts by unique source IP",
    "Write bag of packet counts by unique source IP",
    "Write bag of byte counts by unique source IP",
    "Write bag of flow counts by unique destination IP",
    "Write bag of packet counts by unique destination IP",
    "Write bag of byte counts by unique destination IP",
    "Write bag of flow counts by unique source port",
    "Write bag of packet counts by unique source port",
    "Write bag of byte counts by unique source port",
    "Write bag of flow counts by unique destination port",
    "Write bag of packet counts by unique destination port",
    "Write bag of byte counts by unique destination port",
    "Write bag of flow counts by unique protocol",
    "Write bag of packet counts by unique protocol",
    "Write bag of byte counts by unique protocol",
    (char *)NULL
};



/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  This function is
 *    passed to optionsSetup(); optionsParse() will call this funciton
 *    and then exit the program when the --help option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                        \
    ("[SWITCHES] [FILES]\n"                                              \
     "\tRead SiLK Flow records and builds binary Bag(s) containing\n"    \
     "\tkey-count pairs.  Key is protocol or source or destination\n"    \
     "\taddress or port.  Counter is sum of flows, packets, or bytes.\n" \
     "\tReads SiLK Flows from named files or from the standard input.\n")

    FILE *fh = USAGE_FH;
    int i;

    fprintf(fh, "%s %s", skAppName(), USAGE_MSG);
    fprintf(fh, "\nSWITCHES:\n");
    skOptionsDefaultUsage(fh);
    sksiteCompmethodOptionsUsage(fh);

    /* Loop over the standard options */
    for (i = 0; appOptions[i].name; ++i) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }

    /* Now, print the help for the deprecated options */
    for (i = 0; appOptions[i].name && oldOptions[i].name; ++i) {
        fprintf(fh, "--%s %s. Deprecated alias for --%s\n", oldOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appOptions[i].name);
    }
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;
    int i;
    int rv;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    /* close all bag files */
    for (i = 0; i < NUM_OUTPUTS; ++i) {
        if (bag_io[i].bag) {
            skBag_free(bag_io[i].bag);
        }
        if (bag_io[i].stream) {
            rv = skStreamClose(bag_io[i].stream);
            if (rv) {
                skStreamPrintLastErr(bag_io[i].stream, rv, &skAppPrintErr);
            }
            rv = skStreamDestroy(&bag_io[i].stream);
        }
        memset(&bag_io[i], 0, sizeof(bagfile_t));
    }

    /* close all array bag files */
    for (i = 0; i < NUM_ARRAY_OUTPUTS; ++i) {
        if (array_io[i].array) {
            free(array_io[i].array);
        }
        if (array_io[i].stream) {
            rv = skStreamClose(array_io[i].stream);
            if (rv) {
                skStreamPrintLastErr(array_io[i].stream, rv, &skAppPrintErr);
            }
            rv = skStreamDestroy(&array_io[i].stream);
        }
        memset(&array_io[i], 0, sizeof(arrayfile_t));
    }

    optionsTeardown();
    return;
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    int requested_output;
    int i;
    int rv;

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    skAppRegister(argv[0]);

    /* initialize variables */
    memset(bag_io, 0, sizeof(bag_io));
    memset(array_io, 0, sizeof(array_io));
    memset(overflow_warning, 0, sizeof(overflow_warning));
    memset(array_overflow_warning, 0, sizeof(array_overflow_warning));

    /* initialize options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);
    }

    /* register the options handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)
        || optionsRegister(oldOptions, (optHandler)appOptionsHandler, NULL))
    {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);
    }
    if (sksiteCompmethodOptionsRegister(&comp_method)) {
        skAppPrintErr("unable to register compression option");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    first_file = optionsParse(argc, argv);
    assert(first_file <= argc);
    if (first_file < 0) {
        /* options parsing should print error */
        skAppUsage();           /* never returns */
    }

    /* first_file is looking at first file name to process */
    if (first_file == argc) {
        if (FILEIsATty(stdin)) {
            skAppPrintErr("No input files on command line and"
                          " stdin is connected to a terminal");
            skAppUsage();
        }
    }

    /* verify that the user requested output */
    requested_output = 0;
    for (i = 0; i < NUM_OUTPUTS; ++i) {
        if (bag_io[i].stream) {
            requested_output = 1;
            break;
        }
    }
    if (requested_output == 0) {
        for (i = 0; i < NUM_ARRAY_OUTPUTS; ++i) {
            if (array_io[i].stream) {
                requested_output = 1;
                break;
            }
        }
    }
    if (requested_output == 0) {
        skAppPrintErr("Must specify type of output(s) to generate.");
        skAppUsage();
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    /* open the output files */
    for (i = 0; i < NUM_OUTPUTS; ++i) {
        if (bag_io[i].stream) {
            rv = skStreamOpen(bag_io[i].stream);
            if (rv) {
                skStreamPrintLastErr(bag_io[i].stream, rv, &skAppPrintErr);
                exit(EXIT_FAILURE);
            }
        }
    }
    for (i = 0; i < NUM_ARRAY_OUTPUTS; ++i) {
        if (array_io[i].stream) {
            rv = skStreamOpen(array_io[i].stream);
            if (rv) {
                skStreamPrintLastErr(array_io[i].stream, rv, &skAppPrintErr);
                exit(EXIT_FAILURE);
            }
        }
    }

    return;                     /* OK */
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    This function is passed to optionsRegister(); it will be called
 *    by optionsParse() for each user-specified switch that the
 *    application has registered; it should handle the switch as
 *    required---typically by setting global variables---and return 1
 *    if the switch processing failed or 0 if it succeeded.  Returning
 *    a non-zero from from the handler causes optionsParse() to return
 *    a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    const char *opt_arg)
{
    switch ((appOptionsEnum)opt_index) {
      case SIP_FLOWS:
      case SIP_PKTS:
      case SIP_BYTES:
      case DIP_FLOWS:
      case DIP_PKTS:
      case DIP_BYTES:
        if (!opt_arg || !opt_arg[0]) {
            skAppPrintErr("missing file name for --%s option",
                          appOptions[opt_index].name);
            return 1;
        }
        if (bag_io[opt_index].bag) {
            skAppPrintErr("Option %s given multiple times",
                          appOptions[opt_index].name);
            return 1;
        }
        if (prepareBagFile(&bag_io[opt_index], opt_arg)) {
            return 1;
        }
        break;

      case SPORT_FLOWS:
      case SPORT_PKTS:
      case SPORT_BYTES:
      case DPORT_FLOWS:
      case DPORT_PKTS:
      case DPORT_BYTES:
      case PROTO_FLOWS:
      case PROTO_PKTS:
      case PROTO_BYTES:
          if (!opt_arg || 0 == strlen(opt_arg)) {
              skAppPrintErr("missing file name for --%s option",
                            appOptions[opt_index].name);
              return 1;
          }
          if (array_io[opt_index - AB_OFFSET].array) {
              skAppPrintErr("Option %s given multiple times",
                            appOptions[opt_index].name);
              return 1;
          }
          if ((opt_index == PROTO_FLOWS)
              || (opt_index == PROTO_PKTS)
              || (opt_index == PROTO_BYTES))
          {
              array_io[opt_index - AB_OFFSET].size = 0x100;
          } else {
              array_io[opt_index - AB_OFFSET].size = 0x10000;
          }
          if (prepareArrayBagFile(&array_io[opt_index - AB_OFFSET], opt_arg)) {
              return 1;
          }
          break;
    }

    return 0;                   /* OK */
}


/*
 *  ok = prepareBagFile(bagf, pathname);
 *
 *    Prepare the output bag file 'bagf' to write a bag to 'pathname'.
 *
 *    This function creates an skstream_t to 'pathname' and allocates
 *    a bag that will be writing to that file.  The function makes
 *    sure a file with that name does not currently exist.  If
 *    pathname is "stdout" and no other bag files are writing to
 *    stdout, then stdout will be used.
 *
 *    Returns 0 on success.  Returns non-zero if allocation files, if
 *    we attempt to open an existing file, if more than one bag use
 *    stdout.
 */
static int prepareBagFile(bagfile_t *bagf, const char *pathname)
{
    int rv;

    if (0 == strcmp("stdout", pathname)) {
        if (stdout_used) {
            skAppPrintErr("Only one output may use stdout");
            return 1;
        }
        stdout_used = 1;
    } else if (fileExists(pathname)) {
        skAppPrintErr("Will not overwrite existing file %s", pathname);
        return 1;
    }

    if (skBag_create(&bagf->bag) != SKBAG_OK) {
        skAppPrintErr("Error allocating bag tree");
        return 1;
    }

    if ((rv = skStreamCreate(&(bagf->stream), SK_IO_WRITE, SK_CONTENT_SILK))
        || (rv = skStreamBind(bagf->stream, pathname))
        || (rv = skStreamSetCompressionMethod(bagf->stream, comp_method)))
    {
        skStreamPrintLastErr(bagf->stream, rv, &skAppPrintErr);
        skStreamDestroy(&bagf->stream);
        return 1;
    }

    return 0; /* OK */
}


/*
 *  ok = prepareArrayBagFile(arraybag, pathname);
 *
 *    Prepare the output bag file 'arraybag', which is an in-core
 *    array---of ports or protocols.  The array size should already be
 *    filled in.
 *
 *    This function creates an skstream_t to 'pathname' and allocates
 *    an array that will be written to that file.  The function makes
 *    sure a file with that name does not currently exist.  if
 *    pathname is "stdout" and no other files are writing to stdout,
 *    then stdout will be used.
 *
 *    Returns 0 on success.  Returns non-zero if allocation fails, if
 *    we attempt to open an existing file, if more than one bag uses
 *    stdout.
 */
static int prepareArrayBagFile(arrayfile_t *arrayf, const char *pathname)
{
    int rv;

    assert(arrayf->size);

    if (0 == strcmp("stdout", pathname)) {
        if (stdout_used) {
            skAppPrintErr("Only one output may use stdout");
            return 1;
        }
        stdout_used = 1;
    } else if (fileExists(pathname)) {
        skAppPrintErr("Will not overwrite existing file %s",
                      pathname);
        return 1;
    }

    arrayf->array = calloc(arrayf->size, sizeof(skBag_counter_t));
    if (NULL == arrayf->array) {
        skAppPrintErr("Error allocating bag array");
        return 1;
    }

    if ((rv = skStreamCreate(&(arrayf->stream), SK_IO_WRITE, SK_CONTENT_SILK))
        || (rv = skStreamBind(arrayf->stream, pathname))
        || (rv = skStreamSetCompressionMethod(arrayf->stream, comp_method)))
    {
        skStreamPrintLastErr(arrayf->stream, rv, &skAppPrintErr);
        skStreamDestroy(&arrayf->stream);
        return 1;
    }

    return 0; /* OK */
}


/*
 *  ok = processFile(path);
 *
 *    Read the SiLK Flow records from the file at 'path'---path can be
 *    "stdin"---and potentially create bag files for
 *    {sIP,dIP,sPort,dPort,proto} x {flows,pkts,bytes}.
 *
 *    Return 0 if successful; non-zero otherwise.
 */
static int processFile(const char *path)
{
    const skBag_counter_t max_val = (skBag_counter_t)(~0);
    rwIOStruct *rwIOS;
    skBag_err_t err = SKBAG_OK;
    rwRec rwrec;
    int i;
    int rv = -1; /* return value: assume error */
    skBag_counter_t key;
    skBag_counter_t count;

    rwIOS = rwOpenFile(path, NULL);
    if (!rwIOS) {
        /* rwOpenFile should print error */
        return rv;
    }

    while (rwRead(rwIOS, &rwrec)) {
        for (i = 0; i < NUM_OUTPUTS; ++i) {
            if (bag_io[i].bag == NULL || overflow_warning[i]) {
                continue;
            }

            switch (i) {
              case SIP_FLOWS:
                err = skBag_incrCounter(bag_io[i].bag,
                                        (skBag_key_t*)&rwrec.sIP.ipnum);
                break;

              case DIP_FLOWS:
                err = skBag_incrCounter(bag_io[i].bag,
                                        (skBag_key_t*)&rwrec.dIP.ipnum);
                break;

              case SIP_PKTS:
                count = (skBag_counter_t)rwrec.pkts;
                err = skBag_addToCounter(bag_io[i].bag,
                                         (skBag_key_t*)&rwrec.sIP.ipnum,
                                         &count);
                break;

              case DIP_PKTS:
                count = (skBag_counter_t)rwrec.pkts;
                err = skBag_addToCounter(bag_io[i].bag,
                                         (skBag_key_t*)&rwrec.dIP.ipnum,
                                         &count);
                break;

              case SIP_BYTES:
                count = (skBag_counter_t)rwrec.bytes;
                err = skBag_addToCounter(bag_io[i].bag,
                                         (skBag_key_t*)&rwrec.sIP.ipnum,
                                         &count);
                break;

              case DIP_BYTES:
                count = (skBag_counter_t)rwrec.bytes;
                err = skBag_addToCounter(bag_io[i].bag,
                                         (skBag_key_t*)&rwrec.dIP.ipnum,
                                         &count);
                break;

              default:
                assert(0);
                abort();
            }

            if (err != SKBAG_OK) {
                if (err == SKBAG_ERR_OP_BOUNDS) {
                    overflow_warning[i] = 1;
                    skAppPrintErr("**WARNING** %s overflow for bag %s",
                                  appOptions[i].name,
                                  skStreamGetPathname(bag_io[i].stream));
                } else {
                    skAppPrintErr("Error setting %s value for bag %s: %s",
                                  appOptions[i].name,
                                  skStreamGetPathname(bag_io[i].stream),
                                  skBag_strerror(err));
                    goto END;
                }
            }
        }

        for (i = 0; i < NUM_ARRAY_OUTPUTS; ++i) {
            if (array_io[i].array == NULL || array_overflow_warning[i]) {
                continue;
            }

            switch (i + AB_OFFSET) {
              case SPORT_FLOWS:
                key = (skBag_key_t)rwrec.sPort;
                count = 1;
                break;

              case DPORT_FLOWS:
                key = (skBag_key_t)rwrec.dPort;
                count = 1;
                break;

              case PROTO_FLOWS:
                key = (skBag_key_t)rwrec.proto;
                count = 1;
                break;

              case SPORT_PKTS:
                key = (skBag_key_t)rwrec.sPort;
                count = (skBag_counter_t)rwrec.pkts;
                break;

              case DPORT_PKTS:
                key = (skBag_key_t)rwrec.dPort;
                count = (skBag_counter_t)rwrec.pkts;
                break;

              case PROTO_PKTS:
                key = (skBag_key_t)rwrec.proto;
                count = (skBag_counter_t)rwrec.pkts;
                break;

              case SPORT_BYTES:
                key = (skBag_key_t)rwrec.sPort;
                count = (skBag_counter_t)rwrec.bytes;
                break;

              case DPORT_BYTES:
                key = (skBag_key_t)rwrec.dPort;
                count = (skBag_counter_t)rwrec.bytes;
                break;

              case PROTO_BYTES:
                key = (skBag_key_t)rwrec.proto;
                count = (skBag_counter_t)rwrec.bytes;
                break;

              default:
                assert(0);
                abort();
            }

            if ((max_val - count) < array_io[i].array[key]) {
                array_overflow_warning[i] = 1;
                skAppPrintErr("**WARNING** %s overflow for bag %s",
                              appOptions[i+AB_OFFSET].name,
                              skStreamGetPathname(array_io[i].stream));
                array_io[i].array[key] = max_val;
            } else {
                array_io[i].array[key] += count;
            }
        }
    }

    /* Successful if we make it here */
    rv = 0;

  END:
    rwCloseFile(rwIOS);
    return rv;
}


int main(int argc, char **argv)
{
    int i;

    appSetup(argc, argv);                       /* never returns on error */

    if (first_file == argc) {
        if (0 != processFile("stdin")) {
            return 1;
        }
    } else {
        for (i = first_file; i < argc; ++i) {
            if (0 != processFile(argv[i])) {
                skAppPrintErr("Error processing input file %s",
                              argv[i]);
                return 1;
            }
        }
    }

    /* write out the IP bags */
    for (i = 0; i < NUM_OUTPUTS; ++i) {
        if (bag_io[i].bag) {
            if (SKBAG_OK != skBag_writeBinary(bag_io[i].bag, bag_io[i].stream))
            {
                skAppPrintErr("Error writing to %s",
                              skStreamGetPathname(bag_io[i].stream));
            }
        }
    }

    /* write out the port- and proto-bags */
    for (i = 0; i < NUM_ARRAY_OUTPUTS; ++i) {
        if (array_io[i].array) {
            if (0 != skBag_writeArray(array_io[i].array, array_io[i].size,
                                      array_io[i].stream))
            {
                skAppPrintErr("Error writing to %s",
                              skStreamGetPathname(array_io[i].stream));
            }
        }
    }

    /* done */
    appTeardown();

    return 0;
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
